# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import httpx
import pydantic
from aiopki.lib import JSONWebKeySet
from aiopki.ext.jose import JOSEObject
from starlette.datastructures import URL

from oauthx.lib import utils
from oauthx.lib.exceptions import Error
from oauthx.lib.exceptions import TrustIssues
from .grant import Grant
from .discoverableprovider import DiscoverableProvider
from .lazyservermetadata import LazyServerMetadata
from .obtainedgrant import ObtainedGrant
from .redirectionparameters import RedirectionParameters
from .servermetadata import ServerMetadata


class Provider(pydantic.RootModel[ServerMetadata | LazyServerMetadata | DiscoverableProvider]):
    root: ServerMetadata | LazyServerMetadata | DiscoverableProvider

    @property
    def authorization_endpoint(self) -> URL:
        assert isinstance(self.root, ServerMetadata)
        if not self.root.authorization_endpoint:
            raise NotImplementedError(
                "The authorization server does not provide the "
                "Authorization Endpoint."
            )
        return URL(self.root.authorization_endpoint)

    @property
    def issuer(self) -> str:
        assert isinstance(self.root, ServerMetadata)
        return self.root.issuer

    @property
    def token_endpoint(self) -> URL:
        assert isinstance(self.root, ServerMetadata)
        if not self.root.token_endpoint:
            raise NotImplementedError(
                "The authorization server does not provide the "
                "Token Endpoint."
            )
        return URL(self.root.token_endpoint)

    @property
    def signing_keys(self) -> JSONWebKeySet:
        keys = []
        if isinstance(self.root, ServerMetadata):
            keys = [x for x in self.root.jwks.keys if x.use in {'sig', None}]
        return JSONWebKeySet(keys=keys)

    def authorize(self, **params: Any) -> str:
        return str(self.authorization_endpoint.include_query_params(**params))

    async def discover(self, http: httpx.AsyncClient | None = None):
        if isinstance(self.root, LazyServerMetadata):
            self.root = await self.root.discover(http=http)
        assert isinstance(self.root, ServerMetadata)
        await self.root.discover(http=http)
        return self   

    @utils.http
    async def grant(self, http: httpx.AsyncClient, grant: Grant):
        response = await http.post(
            url=str(self.token_endpoint),
            data=grant.root.model_dump(
                exclude_none=True
            )
        )
        if response.status_code >= 400:
            raise Error(**response.json())
        return ObtainedGrant.model_validate(response.json())

    async def verify(self, params: RedirectionParameters) -> None:
        assert isinstance(self.root, ServerMetadata)
        if self.root.authorization_response_iss_parameter_supported:
            if params.iss is None:
                raise TrustIssues(
                    "The server advertises to support the `iss` parameter "
                    "in authorization responses but did not provide it."
                )

            if params.iss != self.root.issuer:
                raise TrustIssues(
                    f"The `iss` response parameter {params.iss} does not match the "
                    f"expected issuer {self.root.issuer}"
                )

    async def verify_id_token(self, obj: JOSEObject):
        assert not obj.is_encrypted()
        assert isinstance(self.root, ServerMetadata)
        if not await obj.verify(self.root.jwks):
            raise TrustIssues(
                "The signature of the OpenID Connect ID Token did not "
                "validate against the public keys known for issuer "
                f"{self.root.issuer}"
            )

    async def userinfo(self, access_token: str, http: httpx.AsyncClient | None = None):
        assert isinstance(self.root, ServerMetadata)
        return await self.root.userinfo(access_token, http=http)

    def __await__(self):
        return self.discover().__await__()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pytest
import pytest_asyncio
from aiopki.lib import JSONWebKey
from aiopki.lib import JSONWebKeySet
from aiopki.ext.jose import JWE

from oauthx.lib import JARMToken
from .types import AuthorizeFactory


@pytest_asyncio.fixture(scope='function') # type: ignore
async def client_credential(
    rsa_enc: JSONWebKey
) -> Any:
    return JSONWebKeySet(keys=[rsa_enc])


@pytest.fixture(scope='function')
def client_params() -> dict[str, Any]:
    return {
        'authorization_encrypted_response_alg': 'RSA-OAEP-256',
        'authorization_encrypted_response_enc': 'A256GCM'
    }


@pytest.fixture
def response_mode() -> str:
    return 'query.jwt'


@pytest.mark.asyncio
async def test_jarm_uses_client_preferences(authorize_request: AuthorizeFactory):
    raw, *_ = await authorize_request()
    assert isinstance(raw.root, JARMToken)
    assert isinstance(raw.root.response.root, JWE)
    assert raw.root.response.root.alg is not None
    assert raw.root.response.root.enc is not None
    assert raw.root.response.root.alg.name == 'RSA-OAEP-256'
    assert raw.root.response.root.enc.name == 'A256GCM'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import os
import pathlib
from typing import Any

import fastapi
import httpx
import pytest
import pytest_asyncio
import yaml

import aiopki
from aiopki.ext.jose import JWK
from aiopki.ext.keychain import Keychain
from aiopki.utils import b64encode

from oauthx.server.models import ResourceOwner
from oauthx.server import Masker
from oauthx.server import OIDCRouter
from oauthx.server import ObjectFactory
from oauthx.server.config import Config
from oauthx.server.models import Session
from oauthx.server.ref import Storage
from oauthx.server.request import Request
from oauthx.models import Client
from oauthx.types import IStorage


aiopki.install_backend('aiopki.ext.cryptography')


@pytest.fixture(scope='session')
def config_data() -> dict[str, Any]:
    data = yaml.safe_load(open(pathlib.Path(__file__).parent.joinpath('oidc.conf'), 'r').read())
    data['issuer']['id'] = 'https://127.0.0.1'
    return data


@pytest.fixture(scope='function')
def config(config_data: dict[str, Any]) -> Config:
    return Config.model_validate(config_data)


@pytest.fixture(scope='session')
def dek() -> JWK:
    return JWK.model_validate({
        'kty': 'oct',
        'use': 'enc',
        'alg': 'A256GCM',
        'k': os.environ['CONTENT_ENCRYPTION_KEY']
    })


@pytest_asyncio.fixture # type: ignore
async def session_signer(config: Config) -> Any:
    return await config.ui.session_key # type: ignore


@pytest_asyncio.fixture # type: ignore
async def session_cookie(session_signer: Keychain) -> str:
    session = Session.new()
    session.sub = '1'
    return await Request.sign_session(session, session_signer)


@pytest_asyncio.fixture(scope='function') # type: ignore
async def signer(config: Config) -> Keychain:
    await asyncio.gather(*map(asyncio.ensure_future, config.issuer.signing_keys))
    return await Keychain(keys=config.issuer.signing_keys)


@pytest.fixture(scope='function')
def storage_class() -> type[Storage]:
    return type('Storage', (Storage, ), {'objects': {}})


@pytest_asyncio.fixture(autouse=True) # type: ignore
async def storage(
    config: Config,
    storage_class: type[Storage],
    dek: JWK
) -> Storage:
    storage = storage_class(config=config, key=dek)
    factory = ObjectFactory(
        issuer=config.issuer.id,
        storage=storage,
        masker=Masker(
            key=JWK.model_validate({
                'kty': 'oct',
                'use': 'sig',
                'alg': 'HS384',
                'k': os.environ['SECRET_KEY']
            })
        )
    )
    await storage.persist(
        ResourceOwner.model_validate({
            'client_id': 'default',
            'sub': '1',
            'sector_identifier': 'localhost'
        }),
    )
    subject = await factory.subject(pk=1)
    await subject.encrypt_keys(dek)
    await storage.persist(subject)
    return storage


@pytest_asyncio.fixture # type: ignore
async def app(
    config: Config,
    storage_class: type[Storage]
) -> fastapi.FastAPI:
    app = fastapi.FastAPI()
    router = OIDCRouter.from_config(
        app=app,
        config=config,
        prefix='/oauth/v2',
        storage_class=storage_class
    )
    app.include_router(router)
    return app


@pytest_asyncio.fixture(scope='function') # type: ignore
async def agent(app: fastapi.FastAPI, session_cookie: str):
    params = {
        'cookies': {'sessionid':  session_cookie}
    }
    async with httpx.AsyncClient(app=app, **params) as client:
        yield client


@pytest_asyncio.fixture(scope='function') # type: ignore
async def http(app: fastapi.FastAPI):
    async with httpx.AsyncClient(app=app) as client:
        yield client


@pytest_asyncio.fixture(scope='function') # type: ignore
async def client_credential() -> Any:
    return None


@pytest.fixture(scope='function')
def scope() -> list[str]:
    return ['email']


@pytest.fixture(scope='function')
def grant_types() -> list[str]:
    return ['authorization_code']


@pytest.fixture(scope='function')
def response_mode() -> str:
    return 'query'


@pytest.fixture(scope='function')
def response_types() -> list[str]:
    return ['code']


@pytest_asyncio.fixture(scope='function') # type: ignore
async def client(
    client_credential: Any,
    http: httpx.AsyncClient,
    storage: IStorage,
    grant_types: list[str],
    response_types: list[str],
    scope: list[str]
) -> Client:
    client = Client.model_validate({
        'client_id': 'foo',
        'client_name': 'foo',
        'grant_types': grant_types,
        'scope': scope,
        'redirect_uris': ['http://127.0.0.1:23390'],
        'response_types': response_types,
        'credential': client_credential,
        'access_tokens': {
            'claims': ['email'],
            'ttl': 600
        }
    })
    await client.persist(storage)
    return Client.model_validate({
        'provider': {
            'issuer': 'https://127.0.0.1',
            'authorization_endpoint': 'http://localhost/oauth/v2/authorize',
            'token_endpoint': 'http://localhost/oauth/v2/token',
            'authorization_response_iss_parameter_supported': True
        },
        'client_id': 'foo',
        'credential': client_credential
    })


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from oauthx.server.protocols import IResourceOwner
from .authorizationcode import AuthorizationCode
from .subjectkey import SubjectKey


class GrantedAuthorizationCode(pydantic.BaseModel):
    client_id :str
    sub: SubjectKey
    consumed: bool = False
    redirect_uri: str | None
    scope: set[str]
    value: AuthorizationCode

    @property
    def owner(self) -> IResourceOwner.Key:
        return IResourceOwner.Key(self.client_id, str(self.sub))

    def allows_redirect(self, redirect_uri: str) -> bool:
        return any([
            self.redirect_uri is None,
            str(self.redirect_uri) == redirect_uri
        ])

    def consume(self) -> None:
        self.consumed = True

    def has_consent(self, owner: IResourceOwner) -> bool:
        return True

    def is_authorized(self, client_id: str) -> bool:
        return self.client_id == client_id

    def is_consumed(self) -> bool:
        return self.consumed


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.params import DefaultCache
from oauthx.resource.models import Request
from oauthx.resource.models import TokenSubject
from .accesstoken import AccessToken
from .requestcredentials import RequestCredentials


async def get(
    request: Request,
    at: AccessToken,
    cache: DefaultCache,
    credentials: RequestCredentials
) -> TokenSubject:
    subject = await cache.get(request.at_hash, decoder=TokenSubject.model_validate_json)
    if subject is None:
        userinfo = await request.issuer.userinfo(credentials)
        subject = TokenSubject.model_validate({
            'claims': userinfo.model_dump(
                exclude={'aud', 'iss', 'exp', 'iat', 'nbf', 'sub'},
                exclude_none=True
            ),
            'sub': at.sub,
            'at_hash': request.at_hash
        })
        await cache.set(request.at_hash, subject.model_dump_json(), encoder=str.encode)
    return subject


RequestSubject: TypeAlias = Annotated[TokenSubject, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .fatalerror import FatalError


class UnsupportedResponseType(FatalError, ValueError):
    __module__: str = 'oauthx.types'

    def __init__(self, url: str | None = None, allow_redirect: bool = False):
        FatalError.__init__(
            self,
            error='unsupported_response_type',
            error_description=(
                "The requested response type is not allowed by the client or "
                "not supported by the authorization server."
            ),
            error_uri=url,
            allow_redirect=allow_redirect
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import inspect
from typing import Any

import fastapi

from oauthx.lib import utils
from oauthx.lib.exceptions import StopSnooping
from oauthx.lib.params import ClientAuthorizationState
from .models import Client


class ClientResolver:
    __module__: str = 'oauthx.client'
    _is_coroutine = asyncio.coroutines._is_coroutine # type: ignore
    request_attname: str = 'oauth_client'

    @utils.class_property
    def __signature__(cls) -> inspect.Signature:
        return utils.merge_signatures([
            inspect.signature(cls.__call__),
            inspect.signature(cls.resolve)
        ])

    def __new__(cls, *args: Any, **kwargs: Any):
        self = super().__new__(cls, *args, **kwargs)

        # Check if the asyncio.iscoroutinefunction() call returns
        # True for this object, since it depends on a private
        # symbol.
        assert asyncio.iscoroutinefunction(self) # nosec
        return self

    def as_provider(self) -> Any:
        async def f(request: fastapi.Request, state: ClientAuthorizationState) -> None:
            if state is None:
                raise StopSnooping
            setattr(request, 'provider', await self.get(state.client_id))
        return fastapi.Depends(f)

    async def get(self, client_id: str) -> Client:
        raise NotImplementedError

    async def resolve(self, *args: Any, **kwargs: Any) -> Client:
        raise NotImplementedError

    async def __call__(
        self,
        request: fastapi.Request,
        **kwargs: Any
    ) -> None:
        client = await utils.merged_call(self.resolve, {
            'request': request,
            **kwargs
        })
        setattr(request, self.request_attname, client)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from .authorizationcodegrant import AuthorizationCodeGrant


__all__: list[str] = ['BearerToken']


def get(grant: AuthorizationCodeGrant) -> str:
    return grant.access_token


BearerToken: TypeAlias  = Annotated[str, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from oauthx.lib.exceptions import InteractionException
from oauthx.lib.exceptions import StopSnooping

from .accountissues import AccountIssues
from .authorizationkey import AuthorizationKey
from .authorizationrequestkey import AuthorizationRequestKey
from .claimsetkey import ClaimSetKey
from .iauthorizationrequest import IAuthorizationRequest
from .iauthorizationrouter import IAuthorizationRouter
from .iclaimset import IClaimSet
from .invalidauthorizationrequest import InvalidAuthorizationRequest
from .invalidclient import InvalidClient
from .invalidredirecturi import InvalidRedirectURI
from .invalidresponsetype import InvalidResponseType
from .iprincipalmasker import IPrincipalMasker
from .iresponsemode import IResponseMode
from .loginrequired import LoginRequired
from .missingredirecturi import MissingRedirectURI
from .tokenmac import TokenMAC
from .unauthorizedclient import UnauthorizedClient
from .unusableaccount import UnusableAccount


__all__: list[str] = [
    'AuthorizationKey',
    'AuthorizationRequestKey',
    'AccountIssues',
    'ClaimSetKey',
    'IAuthorizationRequest',
    'IAuthorizationRouter',
    'IClaimSet',
    'InteractionException',
    'InvalidAuthorizationRequest',
    'InvalidClient',
    'InvalidRedirectURI',
    'InvalidResponseType',
    'IPrincipalMasker',
    'IResponseMode',
    'LoginRequired',
    'MissingRedirectURI',
    'StopSnooping',
    'TokenMAC',
    'UnauthorizedClient',
    'UnusableAccount',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
from typing import Any

import fastapi

from oauthx.resource import AccessToken
from oauthx.resource import RequestAccessToken
from oauthx.server.params import ContentEncryptionKey
from oauthx.server.params import HTTPClientDependency
from oauthx.server.params import ObjectFactory
from oauthx.server.request import Request
from ..baserequesthandler import BaseRequestHandler
from .params import Client
from .params import Subject
from .userinforoute import UserInfoRoute


class UserInfoEndpointHandler(BaseRequestHandler):
    __module__: str = 'oauthx.server.handlers'
    methods: list[str] = ['GET']
    name: str = 'oidc.userinfo'
    path: str = '/userinfo'
    dependencies = [
        RequestAccessToken(
            issuers={os.getenv('OAUTH_ISSUER_IDENTIFIER') or 'self'},
            audience='any',
            http_factory=HTTPClientDependency
        )
    ]
    description: str = (
        "Returns Claims about the authenticated End-User. To obtain "
        "the requested Claims about the End-User, the Client makes "
        "a request to the UserInfo Endpoint using an Access Token "
        "obtained through OpenID Connect Authentication. These Claims "
        "are normally represented by a JSON object that contains a "
        "collection of name and value pairs for the Claims."
    )
    response_description: str = (
        "Claims about the authenticated end-user"
    )
    route_class = UserInfoRoute
    status_code: int = 200
    summary: str = "UserInfo Endpoint"

    def setup(
        self,
        *,
        access_token: AccessToken,
        factory: ObjectFactory,
        key: ContentEncryptionKey,
        client: Client,
        subject: Subject,
        **_: Any
    ):
        self.access_token = access_token
        self.factory = factory
        self.client = client
        self.key = key
        self.subject = subject

    async def handle(self, request: Request) -> fastapi.Response:
        await self.subject.decrypt_keys(self.key)
        userinfo = await self.factory.userinfo(
            client=self.client,
            subject=self.subject,
            scope=self.access_token.scope
        )
        return fastapi.responses.PlainTextResponse(
            status_code=200,
            media_type='application/json;indent=2',
            content=userinfo.model_dump_json(indent=2, exclude_none=True)
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from typing import Any
from typing import TypeVar

from canonical import ResourceIdentifier

from oauthx.lib.models import ClientAuthorizationStateKey
from oauthx.lib.models import ClientAuthorizationState
from oauthx.lib.protocols import IClientStorage
from oauthx.lib.protocols import IClient
from oauthx.lib.types import Masked
from .datastorestorage import DatastoreStorage
from .params import DatastoreClient


T = TypeVar('T')


class ClientStorage(DatastoreStorage, IClientStorage):

    def __init__( # type: ignore
        self,
        *,
        client: DatastoreClient, # type: ignore
    ) -> None:
        self.client = client

    async def get(self, key: ResourceIdentifier[Any, T]) -> T | None:
        return await self._get(key)

    @functools.singledispatchmethod
    async def _get(self, key: Any) -> Any | None:
        raise NotImplementedError

    async def application(self, client_id: str) -> IClient | None:
        raise NotImplementedError

    async def mask(self, obj: Any) -> Masked:
        raise NotImplementedError

    @_get.register
    async def state(self, key: ClientAuthorizationStateKey) -> Any:
        return await self.get_model_by_key(ClientAuthorizationState, key)

    async def persist(self, object: Any) -> None:
        await self.put(object, object.pk)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.params import ClientAuthorizationState
from oauthx.server.config import Config
from oauthx.server.config import ProviderConfig
from oauthx.server.types import StopSnooping
from .currentconfig import CurrentConfig


__all__: list[str] = ['UpstreamProvider']


async def get(
    state: ClientAuthorizationState,
    config: Config = CurrentConfig,
) -> ProviderConfig:
    if state is None:
        raise StopSnooping
    name = state.annotation('provider')
    if name is None:
        raise NotImplementedError
    provider = config.get_provider(name)
    if provider is None:
        raise NotImplementedError(name)
    return provider


UpstreamProvider: TypeAlias = Annotated[
    ProviderConfig,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from typing import Literal

from .baseclient import BaseClient
from oauthx.lib.models.grant import Grant


class PublicClient(BaseClient):
    credential: Literal[None] = None
    token_endpoint_auth_method: Literal['none'] = 'none'

    @functools.singledispatchmethod
    async def authenticate(self, grant: Grant) -> bool:
        raise NotImplementedError

    def can_encrypt(self) -> bool:
        return False

    def is_confidential(self) -> bool:
        return False


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi
import pydantic

from oauthx.server.request import Request
from oauthx.server.params import IssuerIdentifier
from oauthx.server.params import ObjectFactory
from oauthx.server.types import InvalidAuthorizationRequest
from oauthx.server.types import IResponseMode
from .client import Client
from .query import RESPONSE_MODE
from .redirecturi import RedirectURI
from .responsetype import ResponseType
from .state import State


__all__: list[str] = [
    'ResponseMode'
]


async def get(
    request: Request,
    iss: IssuerIdentifier,
    client: Client,
    response_type: ResponseType,
    redirect_uri: RedirectURI,
    state: State,
    factory: ObjectFactory,
    response_mode: str | None = RESPONSE_MODE,
) -> IResponseMode:
    try:
        obj = await factory.response_mode(
            iss=iss,
            client=client,
            response_type=response_type,
            response_mode=response_mode,
            redirect_uri=redirect_uri,
            state=state
        )
        request.response_mode = obj
        return obj
    except pydantic.ValidationError:
        raise InvalidAuthorizationRequest(
            error='invalid_request',
            allow_redirect=True,
            redirect_uri=redirect_uri,
            context={'client_name': client.get_display_name()}
        )


ResponseMode: TypeAlias = Annotated[
    IResponseMode,
    fastapi.Depends(get)
]